"""App Tasks"""

# Standard Library
import inspect
from collections.abc import Callable

# Third Party
from celery import chain, shared_task

# Django
from django.db.models import Min
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership, User
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, app_settings
from ledger.decorators import when_esi_is_available
from ledger.helpers import data_exporter
from ledger.helpers.discord import send_user_notification
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.models.planetary import CharacterPlanetDetails

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.LEDGER_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

_update_ledger_char_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["character_pk", "force_refresh"], "graceful": True}},
}

_update_ledger_corp_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["corporation_pk", "force_refresh"], "graceful": True}},
}


# pylint: disable=unused-argument, too-many-locals
@shared_task(**TASK_DEFAULTS_ONCE)
def check_planetary_alarms(runs: int = 0):
    all_planets = CharacterPlanetDetails.objects.all()
    owner_ids = {}
    warnings = {}

    for planet in all_planets:
        if planet.is_expired and not planet.notification_sent and planet.notification:
            character_id = planet.planet.character.eve_character.character_id

            try:
                owner = CharacterOwnership.objects.get(
                    character__character_id=character_id
                )
                main = owner.user.profile.main_character
                alts = main.character_ownership.user.character_ownerships.all()

                owner_ids[main.character_id] = alts.values_list(
                    "character__character_id", flat=True
                )

                main_id = main.character_id
            except CharacterOwnership.DoesNotExist:
                continue
            except AttributeError:
                continue

            msg = _("%(charname)s on %(planetname)s") % {
                "charname": planet.planet.character.eve_character.character_name,
                "planetname": planet.planet.planet.name,
            }

            if main_id not in warnings:
                warnings[main_id] = []

            warnings[main_id].append(msg)
            planet.notification_sent = True
            planet.save()

    if warnings:
        for main_id, messages in warnings.items():
            owner = CharacterOwnership.objects.get(character__character_id=main_id)
            title = _("Planetary Extractor Heads Expired")

            # Split messages into chunks of 50
            for i in range(0, len(messages), 50):
                chunk = messages[i : i + 50]
                msg = "\n".join(chunk)

                full_message = format_html(msg)

                send_user_notification.delay(
                    user_id=owner.user.id,
                    title=title,
                    message=full_message,
                    embed_message=True,
                    level="warning",
                )
                runs = runs + 1
    logger.info("Queued %s Planetary Alarms.", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_characters(runs: int = 0, force_refresh=False):
    """Update all characters"""
    # Disable characters with no owner
    CharacterAudit.objects.disable_characters_with_no_owner()
    characters = CharacterAudit.objects.select_related("eve_character").filter(active=1)
    for char in characters:
        update_character.apply_async(
            args=[char.pk], kwargs={"force_refresh": force_refresh}
        )
        runs = runs + 1
    logger.debug("Queued %s Character Audit Tasks", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_subset_characters(subset=2, min_runs=50, max_runs=500, force_refresh=False):
    """Update a batch of characters to prevent overload ESI"""
    # Disable characters with no owner
    CharacterAudit.objects.disable_characters_with_no_owner()
    total_characters = CharacterAudit.objects.filter(active=1).count()
    characters_count = min(max(total_characters // subset, min_runs), total_characters)

    # Limit the number of characters to update to prevent overload ESI
    characters_count = min(characters_count, max_runs)

    # Annotate characters with the oldest `last_run_finished` across all update sections
    characters = (
        CharacterAudit.objects.filter(active=1)
        .annotate(oldest_update=Min("ledger_update_status__last_run_finished_at"))
        .order_by("oldest_update")
        .distinct()[:characters_count]
    )

    for char in characters:
        update_character.apply_async(
            args=[char.pk], kwargs={"force_refresh": force_refresh}
        )
    logger.debug("Queued %s Character Audit Tasks", len(characters))


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_character(character_pk: int, force_refresh=False):
    character = CharacterAudit.objects.prefetch_related("ledger_update_status").get(
        pk=character_pk
    )

    que = []
    priority = 7

    logger.debug(
        "Processing Audit Updates for %s",
        format(character.eve_character.character_name),
    )

    if force_refresh:
        # Reset Token Error if we are forcing a refresh
        character.reset_has_token_error()

    needs_update = character.calc_update_needed()

    if not needs_update and not force_refresh:
        logger.info("No updates needed for %s", character.eve_character.character_name)
        return

    sections = character.UpdateSection.get_sections()

    for section in sections:
        # Skip sections that are not in the needs_update list
        if not force_refresh and not needs_update.for_section(section):
            logger.debug(
                "No updates needed for %s (%s)",
                character.eve_character.character_name,
                section,
            )
            continue

        task_name = f"update_char_{section}"
        task = globals().get(task_name)
        que.append(
            task.si(character.pk, force_refresh=force_refresh).set(priority=priority)
        )

    chain(que).apply_async()
    logger.debug(
        "Queued %s Audit Updates for %s",
        len(que),
        character.eve_character.character_name,
    )


@shared_task(**_update_ledger_char_params)
@when_esi_is_available
def update_char_wallet_journal(character_pk: int, force_refresh: bool):
    return _update_character_section(
        character_pk,
        section=CharacterAudit.UpdateSection.WALLET_JOURNAL,
        force_refresh=force_refresh,
    )


@shared_task(**_update_ledger_char_params)
@when_esi_is_available
def update_char_mining_ledger(character_pk: int, force_refresh: bool):
    return _update_character_section(
        character_pk,
        section=CharacterAudit.UpdateSection.MINING_LEDGER,
        force_refresh=force_refresh,
    )


@shared_task(**_update_ledger_char_params)
@when_esi_is_available
def update_char_planets(character_pk: int, force_refresh: bool):
    logger.debug("Updating Planet Data for %s", character_pk)
    return _update_character_section(
        character_pk,
        section=CharacterAudit.UpdateSection.PLANETS,
        force_refresh=force_refresh,
    )


@shared_task(**_update_ledger_char_params)
@when_esi_is_available
def update_char_planets_details(character_pk: int, force_refresh: bool):
    logger.debug("Updating Planet Details for %s", character_pk)
    return _update_character_section(
        character_pk,
        section=CharacterAudit.UpdateSection.PLANETS_DETAILS,
        force_refresh=force_refresh,
    )


def _update_character_section(character_pk: int, section: str, force_refresh: bool):
    """Update a specific section of the character audit."""
    section = CharacterAudit.UpdateSection(section)
    character = CharacterAudit.objects.get(pk=character_pk)
    logger.debug(
        "Updating %s for %s", section.label, character.eve_character.character_name
    )

    character.reset_update_status(section)

    method: Callable = getattr(character, section.method_name)
    method_signature = inspect.signature(method)

    if "force_refresh" in method_signature.parameters:
        kwargs = {"force_refresh": force_refresh}
    else:
        kwargs = {}
    result = character.perform_update_status(section, method, **kwargs)
    character.update_section_log(section, result)


# Corporation Audit - Tasks
@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_corporations(runs: int = 0, force_refresh=False):
    corps = CorporationAudit.objects.select_related("corporation").filter(active=1)
    for corp in corps:
        update_corporation.apply_async(
            args=[corp.pk], kwargs={"force_refresh": force_refresh}
        )
        runs = runs + 1
    logger.info("Queued %s Corporation Audit Tasks", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_subset_corporations(
    subset=5, min_runs=20, max_runs=200, force_refresh=False
):
    """Update a batch of corporations to prevent overload ESI"""
    total_corporations = CorporationAudit.objects.filter(active=1).count()
    corporations_count = min(
        max(total_corporations // subset, min_runs), total_corporations
    )

    # Limit the number of corporations to update to prevent overload ESI
    corporations_count = min(corporations_count, max_runs)

    # Annotate corporations with the oldest `last_run_finished` across all update sections
    corporations = (
        CorporationAudit.objects.filter(active=1)
        .annotate(
            oldest_update=Min("ledger_corporation_update_status__last_run_finished_at")
        )
        .order_by("oldest_update")
        .distinct()[:corporations_count]
    )

    for corp in corporations:
        update_corporation.apply_async(
            args=[corp.pk], kwargs={"force_refresh": force_refresh}
        )
    logger.debug("Queued %s Corporation Audit Tasks", len(corporations))


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_corporation(
    corporation_pk, force_refresh=False
):  # pylint: disable=unused-argument
    corporation = CorporationAudit.objects.prefetch_related(
        "ledger_corporation_update_status"
    ).get(pk=corporation_pk)

    que = []
    priority = 7

    logger.debug(
        "Processing Audit Updates for %s",
        format(corporation.corporation.corporation_name),
    )

    if force_refresh:
        # Reset Token Error if we are forcing a refresh
        corporation.reset_has_token_error()

    needs_update = corporation.calc_update_needed()

    if not needs_update and not force_refresh:
        logger.info(
            "No updates needed for %s", corporation.corporation.corporation_name
        )
        return

    sections = corporation.UpdateSection.get_sections()

    for section in sections:
        # Skip sections that are not in the needs_update list
        if not force_refresh and not needs_update.for_section(section):
            logger.debug(
                "No updates needed for %s (%s)",
                corporation.corporation.corporation_name,
                section,
            )
            continue

        task_name = f"update_corp_{section}"
        task = globals().get(task_name)
        que.append(
            task.si(corporation.pk, force_refresh=force_refresh).set(priority=priority)
        )

    chain(que).apply_async()
    logger.debug(
        "Queued %s Audit Updates for %s",
        len(que),
        corporation.corporation.corporation_name,
    )


@shared_task(**_update_ledger_corp_params)
@when_esi_is_available
def update_corp_wallet_division_names(corporation_pk: int, force_refresh: bool):
    return _update_corporation_section(
        corporation_pk,
        section=CorporationAudit.UpdateSection.WALLET_DIVISION_NAMES,
        force_refresh=force_refresh,
    )


@shared_task(**_update_ledger_corp_params)
@when_esi_is_available
def update_corp_wallet_division(corporation_pk: int, force_refresh: bool):
    return _update_corporation_section(
        corporation_pk,
        section=CorporationAudit.UpdateSection.WALLET_DIVISION,
        force_refresh=force_refresh,
    )


@shared_task(**_update_ledger_corp_params)
@when_esi_is_available
def update_corp_wallet_journal(corporation_pk: int, force_refresh: bool):
    return _update_corporation_section(
        corporation_pk,
        section=CorporationAudit.UpdateSection.WALLET_JOURNAL,
        force_refresh=force_refresh,
    )


def _update_corporation_section(corporation_pk: int, section: str, force_refresh: bool):
    """Update a specific section of the character audit."""
    section = CorporationAudit.UpdateSection(section)
    corporation = CorporationAudit.objects.get(pk=corporation_pk)
    logger.debug(
        "Updating %s for %s", section.label, corporation.corporation.corporation_name
    )

    corporation.reset_update_status(section)

    method: Callable = getattr(corporation, section.method_name)
    method_signature = inspect.signature(method)

    if "force_refresh" in method_signature.parameters:
        kwargs = {"force_refresh": force_refresh}
    else:
        kwargs = {}

    result = corporation.perform_update_status(section, method, **kwargs)
    corporation.update_section_log(section, result)


@shared_task(**TASK_DEFAULTS_ONCE)
def clear_all_etags():
    logger.debug("Clearing all etags")
    try:
        # Third Party
        # pylint: disable=import-outside-toplevel
        from django_redis import get_redis_connection

        _client = get_redis_connection("default")
    except (NotImplementedError, ModuleNotFoundError):
        # Django
        # pylint: disable=import-outside-toplevel
        from django.core.cache import caches

        default_cache = caches["default"]
        _client = default_cache.get_master_client()
    keys = _client.keys(f":?:{app_settings.LEDGER_CACHE_KEY}-*")
    logger.info("Deleting %s etag keys", len(keys))
    if keys:
        deleted = _client.delete(*keys)
        logger.info("Deleted %s etag keys", deleted)
    else:
        logger.info("No etag keys to delete")


@shared_task(**TASK_DEFAULTS_ONCE)
# pylint: disable=too-many-positional-arguments
def export_data_ledger(
    user_pk: int,
    ledger_type: str,
    entity_id: int,
    division_id: int = None,
    year: int = None,
    month: int = None,
):
    """Export data for a ledger."""
    msg = data_exporter.export_ledger_to_archive(
        ledger_type, entity_id, division_id, year, month
    )

    try:
        user = User.objects.get(pk=user_pk)
    except User.DoesNotExist:
        logger.debug("User with pk %s does not exist", user_pk)
        return

    if not msg:
        title = _(f"{ledger_type.capitalize()} Data Export Ready")
        message = _(
            f"Your data export for topic {ledger_type} is ready. You can download it from the Data Export page.",
        )
    else:
        title = _(f"{ledger_type.capitalize()} Data Export Failed")
        message = _(
            f"Your data export for topic {ledger_type} has failed with following error: {msg}",
        )

    send_user_notification.delay(
        user_id=user.pk,
        title=title,
        message=message,
        embed_message=True,
        level="info",
    )



# Standard Library
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock, Mock, patch
from zipfile import ZipFile

# Django
from django.core.cache import cache
from django.test import RequestFactory, TestCase, override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.helpers import data_exporter
from ledger.models.general import EveEntity
from ledger.tests.testdata.generate_corporationaudit import (
    create_corporationaudit_from_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.helpers.data_exporter"


class TestCorporationLedgerView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.manage_access",
            ],
        )

        cls.user_alliance, cls.character_ownership_alliance = (
            create_user_from_evecharacter(
                1003,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                    "ledger.manage_access",
                ],
            )
        )

        cls.user_no_permissions, cls.character_ownership = (
            create_user_from_evecharacter(
                1002,
                permissions=[
                    "ledger.basic_access",
                    "ledger.advanced_access",
                    "ledger.manage_access",
                ],
            )
        )
        cls.audit = create_corporationaudit_from_user(cls.user)

        cls.audit_alliance = create_corporationaudit_from_user(cls.user_alliance)

        cls.eve_character_first_party = EveEntity.objects.get(eve_id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )

        cls.division_alliance = create_division(
            corporation=cls.audit_alliance,
            name="MEGA KONTO",
            balance=1000000,
            division_id=1,
        )

        cls.journal_entry = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            context_id=1,
            entry_id=10,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

        cls.journal_entry_alliance = create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division_alliance,
            context_id=2,
            entry_id=20,
            amount=1000,
            balance=2000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2016,
                month=10,
                day=29,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test Journal",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="player_donation",
        )

    def test_int_or_none(self):
        """Test should return integer or None."""
        self.assertEqual(data_exporter.int_or_none(100), 100)
        self.assertIsNone(data_exporter.int_or_none("invalid"))

    def test_file_to_zip(self):
        """Test should create a zip file from given files."""
        with TemporaryDirectory() as tmpdirname_1, TemporaryDirectory() as tmpdirname_2:
            # given
            source_file = Path(tmpdirname_1) / "test.csv"
            with source_file.open("w") as fp:
                fp.write("test file")
            destination = Path(tmpdirname_2)

            zip_file = data_exporter.file_to_zip(source_file, destination=destination)
            with ZipFile(zip_file, "r") as testzip:
                namelist = testzip.namelist()
            self.assertIn(source_file.name, namelist)

    def test_default_destination(self):
        """Test should return default destination path."""
        with TemporaryDirectory() as tmpdirname:
            with patch(
                "ledger.helpers.data_exporter.default_destination",
                return_value=Path(tmpdirname),
            ):
                destination = data_exporter.default_destination()
                self.assertEqual(destination, Path(tmpdirname))

    def test_export_ledger_to_archive_corporation(self):
        """Test should export corporation ledger to archive."""
        with TemporaryDirectory():
            ledger_type = "corporation"

            corporation_id = self.audit.corporation.corporation_id
            year = 2016

            result = data_exporter.export_ledger_to_archive(
                ledger_type=ledger_type,
                entity_id=corporation_id,
                division_id=None,
                year=year,
                month=None,
            )
            # then
            output_file = Path(result)

            self.assertTrue(output_file.exists())
            self.assertEqual(output_file.suffix, ".zip")
            output_file.unlink(missing_ok=True)  # Clean up the created file

    def test_export_ledger_to_archive_alliance(self):
        """Test should export alliance ledger to archive."""
        with TemporaryDirectory():
            ledger_type = "alliance"

            alliance_id = self.audit_alliance.corporation.alliance.alliance_id
            year = 2016

            result = data_exporter.export_ledger_to_archive(
                ledger_type=ledger_type,
                entity_id=alliance_id,
                division_id=None,
                year=year,
                month=None,
            )
            # then
            output_file = Path(result)

            self.assertTrue(output_file.exists())
            self.assertEqual(output_file.suffix, ".zip")
            output_file.unlink(missing_ok=True)  # Clean up the created file



# Standard Library
from unittest.mock import patch

# Django
from django.test import override_settings

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase
from eveuniverse.models import EveSolarSystem, EveType

# AA Ledger
from ledger.tests.testdata.esi_stub import esi_client_stub_openapi
from ledger.tests.testdata.generate_characteraudit import (
    create_characteraudit_from_evecharacter,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.character_mining_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi")
@patch("ledger.models.characteraudit.CharacterMiningLedger.update_evemarket_price")
class TestCharacterMiningManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()
        cls.audit = create_characteraudit_from_evecharacter(1001)

        cls.eve_type = EveType.objects.get(id=17425)
        cls.eve_system = EveSolarSystem.objects.get(id=30004783)

    def test_update_mining_ledger(self, _, __, mock_esi):
        # given
        mock_esi.client = esi_client_stub_openapi
        self.audit.update_mining_ledger(force_refresh=False)

        obj = self.audit.ledger_character_mining.filter(
            date__contains="2014-10-29"
        ).first()
        self.assertEqual(obj.quantity, 5000)
        self.assertEqual(obj.system_id, 30004783)
        self.assertEqual(obj.type_id, 17425)



# Generated by Django 4.2.11 on 2024-12-06 08:37

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0005_alter_corporationaudit_options_alter_general_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="characteraudit",
            name="character_name",
            field=models.CharField(default=None, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="characterplanet",
            name="planet_name",
            field=models.CharField(default=None, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="corporationaudit",
            name="corporation_name",
            field=models.CharField(default=None, max_length=100, null=True),
        ),
    ]



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.eveonline.providers import Alliance, ObjectNotFound

# AA Ledger
from ledger.tests.testdata.generate_corporationaudit import (
    create_user_from_evecharacter,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views.alliance.add_ally import add_ally

MODULE_PATH = "ledger.views.alliance.add_ally"


@patch(MODULE_PATH + ".messages")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestAddAllyView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )
        cls.alliance = Alliance(
            id=3005,
            name="Test Alliance",
            ticker="T.E.S.T",
            corp_ids=[2001, 2002],
            executor_corp_id=None,
            faction_id=None,
        )

    def _add_alliance(self, user, token):
        request = self.factory.get(reverse("ledger:add_ally"))
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = add_ally.__wrapped__.__wrapped__.__wrapped__
        return orig_view(request, token)

    def test_add_ally_already_exist(self, mock_messages):
        # given
        user = self.user
        token = user.token_set.get(character_id=1001)
        # when
        response = self._add_alliance(user, token)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("ledger:alliance_ledger", args=[3001]))
        self.assertEqual(mock_messages.info.call_count, 1)

    @patch(MODULE_PATH + ".provider")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.get_or_create")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.get")
    def test_add_ally_does_not_exist(
        self, mock_get, mock_get_or_create, mock_provider, mock_messages
    ):
        # given
        mock_get.side_effect = EveAllianceInfo.DoesNotExist
        mock_provider.get_alliance.return_value = self.alliance

        mock_ally = Mock()
        mock_ally.populate_alliance = Mock()
        mock_ally.alliance_id = 3005
        mock_ally.alliance_name = "Test Alliance"
        mock_ally.alliance_ticker = "T.E.S.T"
        mock_ally.executor_corp_id = 2001
        mock_get_or_create.return_value = (mock_ally, True)
        user = self.user
        token = user.token_set.get(character_id=1001)
        # when
        response = self._add_alliance(user, token)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("ledger:alliance_ledger", args=[3005]))
        self.assertEqual(mock_messages.success.call_count, 1)
        mock_get.assert_called_once_with(alliance_id=3001)
        mock_provider.get_alliance.assert_called_once_with(3001)
        mock_get_or_create.assert_called_once()

    @patch(MODULE_PATH + ".provider")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.get_or_create")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.get")
    def test_add_ally_does_not_exist_object_not_found(
        self, mock_get, mock_get_or_create, mock_provider, mock_messages
    ):
        # given
        mock_get.side_effect = EveAllianceInfo.DoesNotExist
        mock_provider.get_alliance.side_effect = ObjectNotFound(3001, "alliance")

        user = self.user
        token = user.token_set.get(character_id=1001)
        # when
        response = self._add_alliance(user, token)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("ledger:alliance_ledger", args=[3001]))
        self.assertEqual(mock_messages.warning.call_count, 1)
        mock_get_or_create.assert_not_called()



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import HTTPNotModified

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EvePlanet, EveType

# AA Ledger
from ledger import __title__
from ledger.constants import COMMAND_CENTER, EXTRACTOR_CONTROL_UNIT, SPACEPORTS
from ledger.decorators import log_timing
from ledger.models.characteraudit import CharacterAudit
from ledger.providers import esi

if TYPE_CHECKING:  # pragma: no cover
    # Alliance Auth
    from esi.stubs import CharactersCharacterIdPlanetsGetItem as PlanetGetItem
    from esi.stubs import CharactersCharacterIdPlanetsPlanetIdGet as PlanetDetailsItem

    # AA Ledger
    from ledger.models.general import UpdateSectionResult
    from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def to_json_serializable(data):
    if isinstance(data, dict):
        return {k: to_json_serializable(v) for k, v in data.items()}
    if isinstance(data, list):
        return [to_json_serializable(i) for i in data]
    if hasattr(data, "__dict__"):
        return to_json_serializable(data.__dict__)
    if isinstance(data, timezone.datetime):
        return data.isoformat()
    return data


class PlanetaryQuerySet(models.QuerySet):
    pass


class PlanetaryManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, character: "CharacterAudit", force_refresh: bool = False
    ) -> "UpdateSectionResult":
        """Update or Create a planets entry from ESI data."""
        return character.update_section_if_changed(
            section=character.UpdateSection.PLANETS,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(
        self, audit: "CharacterAudit", force_refresh: bool = False
    ) -> None:
        """Fetch planetary entries from ESI data."""
        req_scopes = ["esi-planets.manage_planets.v1"]
        token = audit.get_token(scopes=req_scopes)

        # Make the ESI request
        operation = esi.client.Planetary_Interaction.GetCharactersCharacterIdPlanets(
            character_id=audit.eve_character.character_id,
            token=token,
        )

        planets_items = operation.results(force_refresh=force_refresh)

        self._update_or_create_objs(character=audit, objs=planets_items)

    @transaction.atomic()
    def _update_or_create_objs(
        self, character: "CharacterAudit", objs: list["PlanetGetItem"]
    ) -> None:
        """Update or Create planets entries from objs data."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.planetary import CharacterPlanetDetails

        _current_planets = self.filter(character=character).values_list(
            "planet_id", flat=True
        )

        _planets_ids = []
        _planets_new = []
        _planets_update = []

        for planet in objs:
            eve_planet, _ = EvePlanet.objects.get_or_create_esi(id=planet.planet_id)
            _planets_ids.append(eve_planet.id)

            try:
                e_planet = self.get(character=character, planet=eve_planet)
                prim_key = e_planet.id
            except self.model.DoesNotExist:
                prim_key = None

            planet = self.model(
                id=prim_key,
                character=character,
                planet=eve_planet,
                planet_name=eve_planet.name,
                upgrade_level=planet.upgrade_level,
                num_pins=planet.num_pins,
            )

            if eve_planet.id not in _current_planets:
                _planets_new.append(planet)
            else:
                _planets_update.append(planet)

        if _planets_new:
            self.bulk_create(_planets_new)
        if _planets_update:
            self.bulk_update(_planets_update, fields=["upgrade_level", "num_pins"])

        # Delete Planets that are no longer in the list
        obsolete_planets = set(_current_planets) - set(_planets_ids)
        self.filter(character=character, planet_id__in=obsolete_planets).delete()
        # Delete Planet Details that are no longer in the list
        CharacterPlanetDetails.objects.filter(
            planet__character=character, planet__planet_id__in=obsolete_planets
        ).delete()


PlanetaryManager = PlanetaryManagerBase.from_queryset(PlanetaryQuerySet)


class PlanetaryDetailsQuerySet(models.QuerySet):
    def get_or_create_facilitys(self, planet: "CharacterPlanet"):
        """Get or Create Facilitys for a given Planet"""
        return self._get_or_create_facilitys(planet=planet)

    def _get_or_create_facilitys(self, planet: "CharacterPlanet"):
        try:
            facilitys = self.get(planet=planet)
            facility = facilitys.facilitys
            created = False
            return facility, created
        except self.model.DoesNotExist:
            facility, created = self.update_or_create_facilitys(planet=planet)
        return facility, created

    def update_or_create_layout(
        self,
        character: CharacterAudit,
        planet: "CharacterPlanetDetails",
        objs: list["PlanetDetailsItem"],
    ):
        """Update or Create Layout for a given Planet"""
        return self._update_or_create(character=character, planet=planet, objs=objs)

    def _convert_to_dict(
        self,
        result: list,
    ) -> tuple:
        objects_list = []
        for data in result:
            data_dict = to_json_serializable(data)
            objects_list.append(data_dict)
        return objects_list

    def _update_or_create(
        self,
        character: "CharacterAudit",
        planet: "CharacterPlanetDetails",
        objs: list["PlanetDetailsItem"],
    ):
        """Update or Create Layout for a given Planet"""
        if not isinstance(objs, list):
            objs = [objs]
        for result in objs:
            links = self._convert_to_dict(result.links)
            pins = self._convert_to_dict(result.pins)
            routes = self._convert_to_dict(result.routes)

            planetdetails, created = self.update_or_create(
                planet=planet,
                character=character,
                defaults={
                    "links": links,
                    "pins": pins,
                    "routes": routes,
                    "facilitys": None,
                },
            )

            self._update_facility(planetdetails)

            logger.debug("Planet %s Facilitys Updated", planetdetails)

            return planetdetails, created

    def _update_facility(self, planet: "CharacterPlanetDetails"):
        facility_info = self.get_facility_info(planet)
        planet.facilitys = facility_info
        planet.save()
        return planet

    # TODO make code easier or split it to peaces for better readable
    def get_facility_info(self, planet: "CharacterPlanetDetails"):
        facility_info = {}

        # Process pins
        for pin in planet.pins:
            pin_id = pin["pin_id"]
            item_type, _ = EveType.objects.get_or_create_esi(id=pin["type_id"])
            if (
                pin["type_id"] in SPACEPORTS
                or pin["type_id"] in EXTRACTOR_CONTROL_UNIT
                or pin["type_id"] in COMMAND_CENTER
            ):
                continue
            facility_info[pin_id] = {
                "facility_id": pin["type_id"],
                "facility_name": item_type.name,
                "resources": [],
                "storage": {
                    content["type_id"]: content["amount"]
                    for content in pin.get("contents", [])
                },
            }

        self._facility_production_chain(planet, facility_info)

        self._facility_produce_depend(planet, facility_info)

        return facility_info

    def _facility_production_chain(
        self, planet: "CharacterPlanetDetails", facility_info: dict
    ):
        """Update dict and add all Production Information to the given dict"""

        extractors = planet.get_extractors_info()
        extractors_item_ids = [
            extractor["item_id"] for extractor in extractors.values()
        ]

        for route in planet.routes:
            destination_pin_id = route["destination_pin_id"]
            source_pin_id = route["source_pin_id"]
            content_type, _ = EveType.objects.get_or_create_esi(
                id=route["content_type_id"]
            )

            if destination_pin_id in facility_info:
                req_quantity = route["quantity"]
                current_quantity = facility_info[destination_pin_id]["storage"].get(
                    content_type.id, 0
                )
                missing_quantity = req_quantity - current_quantity
                still_producing = (
                    content_type.id in extractors_item_ids
                    if not planet.is_expired
                    else False
                )

                resource = {
                    "item_id": content_type.id,
                    "item_name": content_type.name,
                    "req_quantity": req_quantity,
                    "current_quantity": current_quantity,
                    "missing_quantity": max(missing_quantity, 0),
                    "still_producing": still_producing,
                }
                facility_info[destination_pin_id]["resources"].append(resource)

            if source_pin_id in facility_info:
                facility_info[source_pin_id]["output_product"] = {
                    "item_id": content_type.id,
                    "item_name": content_type.name,
                    "output_quantity": route["quantity"],
                }
        return facility_info

    def _facility_produce_depend(self, planet, facility_info):
        """Check if Facility is Producing"""
        for facility in facility_info.values():
            for resource in facility["resources"]:
                self._facility_still_producing(planet, resource, facility_info)
        return facility_info

    def _facility_still_producing(self, planet, resource, facility_info) -> bool:
        """Check if ressource exist in Production Chain and change State"""
        for other_facility in facility_info.values():
            if (
                "output_product" in other_facility
                and other_facility["output_product"]["item_id"] == resource["item_id"]
                and not planet.is_expired
            ):
                if other_facility["output_product"]["output_quantity"] > 0:
                    resource["still_producing"] = True


class PlanetaryDetailsManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, character: "CharacterAudit", force_refresh: bool = False
    ) -> "UpdateSectionResult":
        """Update or Create a planets details entry from ESI data."""
        return character.update_section_if_changed(
            section=character.UpdateSection.PLANETS_DETAILS,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(
        self, audit: "CharacterAudit", force_refresh: bool = False
    ) -> None:
        """Fetch planets details entries from ESI data."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.planetary import CharacterPlanet

        req_scopes = ["esi-planets.manage_planets.v1"]

        token = audit.get_token(scopes=req_scopes)

        planets_ids = CharacterPlanet.objects.filter(character=audit).values_list(
            "planet_id", flat=True
        )
        is_updated = False

        for planet_id in planets_ids:
            # Make the ESI request
            operation = esi.client.Planetary_Interaction.GetCharactersCharacterIdPlanetsPlanetId(
                character_id=audit.eve_character.character_id,
                planet_id=planet_id,
                token=token,
            )

            try:
                planets_details_items = operation.results(force_refresh=force_refresh)
                is_updated = True
            except HTTPNotModified:
                continue

            self._update_or_create_objs(
                character=audit,
                objs=planets_details_items,
                planet_id=planet_id,
            )
        # Raise if no update happened at all
        if not is_updated:
            raise HTTPNotModified(304, {"msg": "Planets Details has Not Modified"})

    @transaction.atomic()
    def _update_or_create_objs(
        self,
        character: "CharacterAudit",
        objs: list["PlanetDetailsItem"],
        planet_id: int,
    ) -> None:
        """Update or Create planets entries from objs data."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.planetary import CharacterPlanet

        try:
            character_planet = CharacterPlanet.objects.get(
                character=character, planet_id=planet_id
            )
        except CharacterPlanet.DoesNotExist:
            logger.warning(
                "Planet %s not found for character %s",
                planet_id,
                character.eve_character.character_name,
            )
            return

        planet_details, created = self.update_or_create_layout(
            character=character,
            planet=character_planet,
            objs=objs,
        )

        if not created:
            # Set Alert if Extractor Heads are expired
            if planet_details.is_expired and planet_details.last_alert is None:
                logger.debug(
                    "Planet %s Extractor Heads Expired for: %s",
                    planet_details.planet.planet.name,
                    planet_details.planet.character.eve_character.character_name,
                )
                planet_details.last_alert = timezone.now()

            # Reset Alert after 1 day
            if (
                planet_details.last_alert is not None
                and planet_details.last_alert
                < timezone.now() - timezone.timedelta(days=1)
            ):
                logger.debug(
                    "Notification Reseted for %s Planet: %s",
                    planet_details.planet.character.eve_character.character_name,
                    planet_details.planet.planet.name,
                )
                planet_details.last_alert = None
                planet_details.notification_sent = False

            planet_details.save()


PlanetaryDetailsManager = PlanetaryDetailsManagerBase.from_queryset(
    PlanetaryDetailsQuerySet
)



"""TestView class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# AA Ledger
from ledger.tests.testdata.generate_corporationaudit import (
    add_corporationaudit_corporation_to_user,
    create_user_from_evecharacter,
)

# AA Skillfarm
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views.corporation.corporation_ledger import corporation_delete

MODULE_PATH = "ledger.views.corporation.corporation_ledger"


class TestDeleteCorporationView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.corp_audit_manager",
                "ledger.manage_access",
            ],
        )
        cls.audit = add_corporationaudit_corporation_to_user(cls.user, 1001)
        cls.no_audit_user, cls.character_ownership = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.manage_access",
            ],
        )

    def test_delete_corporation(self):
        corporation_id = self.audit.corporation.corporation_id

        request = self.factory.post(
            reverse("ledger:delete_corp", args=[corporation_id])
        )
        request.user = self.user

        response = corporation_delete(request, corporation_id=corporation_id)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTrue(response_data["success"])
        self.assertEqual(response_data["message"], "Hell RiderZ successfully deleted")

    def test_delete_corporation_no_audit(self):
        request = self.factory.post(reverse("ledger:delete_corp", args=[2002]))
        request.user = self.user

        response = corporation_delete(request, corporation_id=2002)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Corporation not found")

    def test_delete_corporation_no_permission(self):
        """Test deleting a corporation without being in the corporation."""
        request = self.factory.post(reverse("ledger:delete_corp", args=[2001]))
        request.user = self.no_audit_user

        response = corporation_delete(request, corporation_id=2001)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Permission Denied")



"""Forms for app."""

# Django
from django import forms
from django.utils.translation import gettext_lazy as _

# AA Ledger
from ledger.models.corporationaudit import (
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)


class ConfirmForm(forms.Form):
    """Form Confirms."""

    character_id = forms.CharField(
        widget=forms.HiddenInput(),
    )

    planet_id = forms.CharField(
        widget=forms.HiddenInput(),
    )


class DivisionModelChoiceField(forms.ModelChoiceField):
    """ModelChoiceField that displays "division_id - name" as the option label."""

    def label_from_instance(self, obj: CorporationWalletDivision) -> str:
        name = obj.name or ""
        return f"{name}"


class YearlyModelChoiceField(forms.TypedChoiceField):
    """TypeChoiceField for selecting a year built from date objects."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("coerce", int)
        super().__init__(*args, **kwargs)

    def label_from_instance(self, obj) -> str:
        return f"{obj}"


class GenerateDataExportForm(forms.Form):
    """Form to generate data export."""

    def __init__(self, *args, corporation_id, **kwargs):
        super().__init__(*args, **kwargs)

        # Set the division queryset based on the corporation_id
        self.fields["division"].queryset = CorporationWalletDivision.objects.filter(
            corporation__corporation__corporation_id=corporation_id
        ).order_by("division_id")

        # Populate the year field with distinct years from the journal entries
        years_qs = (
            CorporationWalletJournalEntry.objects.filter(
                division__corporation__corporation__corporation_id=corporation_id
            )
            .exclude(date__year__isnull=True)
            .values_list("date__year", flat=True)
            .order_by("-date__year")
            .distinct()
        )

        year_choices = [("", "---------")] + [(str(y), str(y)) for y in years_qs]
        self.fields["year"].choices = year_choices

    year = YearlyModelChoiceField(
        choices=[],
        required=True,
        widget=forms.Select(attrs={"class": "form-select"}),
    )

    MONTH_CHOICES = [
        ("", _("All months")),
        ("1", _("January")),
        ("2", _("February")),
        ("3", _("March")),
        ("4", _("April")),
        ("5", _("May")),
        ("6", _("June")),
        ("7", _("July")),
        ("8", _("August")),
        ("9", _("September")),
        ("10", _("October")),
        ("11", _("November")),
        ("12", _("December")),
    ]

    month = forms.TypedChoiceField(
        choices=MONTH_CHOICES,
        required=False,
        coerce=lambda val: (
            int(val) if val not in ("", None) else None
        ),  # Coerce to int or None
        empty_value=None,
        widget=forms.Select(attrs={"class": "form-select"}),
    )
    division = DivisionModelChoiceField(
        queryset=CorporationWalletDivision.objects.none(),
        required=False,
        empty_label=_("All Divisions"),
        widget=forms.Select(attrs={"class": "form-select"}),
    )



# Standard Library
import csv
import gc
import tempfile
import zipfile
from abc import ABC, abstractmethod
from decimal import Decimal
from pathlib import Path

# Django
from django.conf import settings
from django.utils import timezone

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.helpers.alliance import AllianceData
from ledger.helpers.corporation import CorporationData
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def int_or_none(val) -> int | None:
    """Convert value to int or return None."""
    if val is None or val == "":
        return None
    try:
        return int(val)
    except (ValueError, TypeError):
        return None


def file_to_zip(source_file: Path, destination: Path) -> Path:
    """Create a zip archive from a file."""
    destination.mkdir(parents=True, exist_ok=True)
    zip_file = (destination / source_file.name).with_suffix(".zip")
    with zipfile.ZipFile(
        file=zip_file, mode="w", compression=zipfile.ZIP_DEFLATED
    ) as my_zip:
        my_zip.write(filename=source_file, arcname=source_file.name)
    logger.info("Created export file: %s", zip_file)
    return zip_file


def default_destination() -> Path:
    """Return default destination path."""
    return (
        Path(settings.BASE_DIR) / str(CorporationAudit._meta.app_label) / "data_exports"
    )


def export_ledger_to_archive(
    ledger_type: str, entity_id: int, division_id: int, year: int, month: int
) -> str:
    """Export data for given ledger into a zipped file in destination."""
    exporter = LedgerCSVExporter.create_exporter(
        ledger_type, entity_id, division_id, year, month
    )
    try:
        with tempfile.TemporaryDirectory() as temp_dirname:
            csv_file = exporter.write_to_file(temp_dirname)
            destination = default_destination()
            zip_file_path = file_to_zip(csv_file, destination)
    except ValueError as ve:
        logger.debug("Error exporting ledger to archive: %s", ve)
        return f"{ve}"
    gc.collect()
    return str(zip_file_path)


class LedgerCSVExporter(ABC):
    """CSV exporter for ledger data."""

    def __init__(self):
        # _now is available for subclasses if needed
        self._now = timezone.now()
        if not hasattr(self, "topic"):
            raise NotImplementedError("Subclasses must define 'topic' attribute")

    @property
    def output_basename(self) -> Path:
        """Return the base name for the output file."""
        app_label = str(CorporationAudit._meta.app_label)
        # Try to discover an entity id from known attributes set by exporters
        entity_id = (
            getattr(self, "corporation_id", None)
            or getattr(self, "alliance_id", None)
            or ""
        )
        division_id = getattr(self, "division_id", None) or ""
        year = getattr(self, "year", "") or ""
        month = getattr(self, "month", "") or ""

        export_key = self.encoder(entity_id, division_id, year, month)
        if not export_key:
            raise ValueError("Could not create export key for output basename")
        # pylint: disable=no-member
        return Path(f"{app_label}_{self.topic}_{export_key}")

    @staticmethod
    def encoder(
        entity_id: int, division_id: int = None, year: int = None, month: int = None
    ) -> str | None:
        """Create a compact export key: "<entity_id>:<division_id>:<year>:<month>" encoded as hex"""
        try:
            parts = [
                "" if v is None else str(v)
                for v in (entity_id, division_id, year, month)
            ]
            key = ":".join(parts)
            return key.encode("utf-8").hex()
        # pylint: disable=broad-except
        except Exception:
            return None

    @staticmethod
    def decoder(key: str) -> tuple[int | None, int | None, int | None, int | None]:
        """Decode a compact export key from hex to (entity_id, division_id, year, month).

        Returns ints when values are numeric, otherwise None for missing/empty values.
        """
        try:
            packed = bytes.fromhex(key).decode("utf-8")
            key_parts = packed.split(":")

            entity_id = int_or_none(key_parts[0]) if len(key_parts) > 0 else None
            division_id = int_or_none(key_parts[1]) if len(key_parts) > 1 else None
            year = int_or_none(key_parts[2]) if len(key_parts) > 2 else None
            month = int_or_none(key_parts[3]) if len(key_parts) > 3 else None
            return entity_id, division_id, year, month
        # pylint: disable=broad-except
        except Exception:
            return None, None, None, None

    @abstractmethod
    def create_data_export(self) -> list[dict]:
        """Generate data export."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def has_data(self) -> bool:
        """Check if there is data to export."""
        raise NotImplementedError()

    def output_path(self, destination: str) -> Path:
        """Return output path for this export."""
        return Path(destination) / self.output_basename.with_suffix(".csv")

    def _extract_entity_info(self, item: dict) -> dict:
        """Extract entity information from the data item."""
        entity = item.get("entity")
        if not entity:
            return {"entity_id": "", "entity_name": "", "entity_type": ""}

        return {
            "entity_id": getattr(entity, "entity_id", ""),
            "entity_name": getattr(entity, "entity_name", ""),
            "entity_type": getattr(entity, "type", ""),
        }

    def _clean_decimal_value(self, value) -> str:
        """Clean decimal values from the Decimal() wrapper."""
        if (
            isinstance(value, str)
            and value.startswith('Decimal("')
            and value.endswith('")')
        ):
            # Extract the value from Decimal("value") format
            return value[9:-2]  # Remove 'Decimal("' and '")')
        if isinstance(value, Decimal):
            return str(value)
        return str(value) if value is not None else "0"

    def _extract_ledger_data(self, item: dict) -> dict:
        """Extract ledger financial data from the item."""
        ledger = item.get("ledger", {})

        return {
            "bounty": self._clean_decimal_value(ledger.get("bounty", "0")),
            "ess": self._clean_decimal_value(ledger.get("ess", "0")),
            "miscellaneous": self._clean_decimal_value(
                ledger.get("miscellaneous", "0")
            ),
            "costs": self._clean_decimal_value(ledger.get("costs", "0")),
            "total": self._clean_decimal_value(ledger.get("total", "0")),
            "mining": self._clean_decimal_value(
                ledger.get("mining", "0")
            ),  # Include mining if present
        }

    def write_to_file(self, destination: Path) -> Path:
        """Write ledger data to CSV file."""
        output_file = self.output_path(destination)

        # Define CSV headers
        headers = [
            "entity_id",
            "entity_name",
            "entity_type",
            "bounty",
            "ess",
            "miscellaneous",
            "costs",
            "mining",
            "total",
        ]

        if self.has_data is False:
            raise ValueError("No data to export")

        with output_file.open("w", newline="", encoding="utf-8") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()

            for item in self.create_data_export():
                if not item:  # Skip empty items
                    continue

                # Extract entity and ledger information
                entity_info = self._extract_entity_info(item)
                ledger_data = self._extract_ledger_data(item)

                # Combine all data into a single row
                row_data = {**entity_info, **ledger_data}

                writer.writerow(row_data)
        logger.info(
            f"Exported {len(self.create_data_export())} ledger entries to {output_file}"
        )
        return output_file

    @classmethod
    # pylint: disable=too-many-positional-arguments
    def create_exporter(
        cls,
        ledger_type: str,
        entity_id: int,
        division_id: int = None,
        year: int = None,
        month: int = None,
    ):
        """Create a LedgerCSVExporter instance."""
        if ledger_type == "corporation":
            if entity_id is None:
                raise ValueError("corporation_id must be provided")
            return CorporationExporter(
                entity_id=entity_id, division_id=division_id, year=year, month=month
            )

        if ledger_type == "alliance":
            if entity_id is None:
                raise ValueError("alliance_id must be provided")
            return AllianceExporter(entity_id=entity_id, year=year, month=month)

        raise ValueError("Invalid ledger_type")

    def gather_export_files(self) -> list[dict]:
        """Gather export files in the destination folder and parse metadata from filenames.
        Expected filename format (without suffix):
            <app_label>_<topic>_<export_key_hex>

        The export_key_hex is a hex encoding of the UTF-8 string
            "<entity_id>:<division_id>:<year>:<month>"
        Only files matching this new format are considered.

        This is an instance method and will filter files for the exporter instance
        by using the instance's corporation_id or alliance_id attribute.
        """
        # Determine the entity id for this exporter instance. Exporter subclasses
        instance_entity_id = getattr(self, "corporation_id", None) or getattr(
            self, "alliance_id", None
        )
        if instance_entity_id is None:
            raise ValueError(
                "Exporter instance must have 'corporation_id' or 'alliance_id' to gather files"
            )

        destination_folder = default_destination()
        results: list[dict] = []

        # If destination folder doesn't exist, return empty
        if not destination_folder.exists():
            return results

        existing_files = list(
            destination_folder.glob(f"{str(CorporationAudit._meta.app_label)}_*.zip")
        )

        if not existing_files:
            return results

        for file in existing_files:
            name = file.with_suffix("").name
            parts = name.split("_")
            topic = parts[1] if len(parts) > 1 else ""

            # Only attempt to parse the export hex key (third segment).
            if len(parts) <= 2:
                continue

            key = parts[2]
            entity_id, division_id, year, month = LedgerCSVExporter.decoder(key)

            # Only include files matching the requested entity id and topic
            if entity_id is None or entity_id != instance_entity_id:
                continue

            # Also filter by exporter topic so we don't mix corporation/alliance exports
            if topic != getattr(self, "topic", ""):
                continue

            try:
                last_updated = timezone.datetime.fromtimestamp(
                    file.stat().st_mtime, tz=timezone.utc
                )
            # pylint: disable=broad-except
            except Exception:
                logger.debug("Could not get last modified time for file %s", file)
                last_updated = None

            results.append(
                {
                    "topic": topic,
                    "entity_id": entity_id,
                    "division_id": (
                        division_id if division_id is not None else "All Divisions"
                    ),
                    "year": year,
                    "month": month if month is not None else "All Months",
                    "last_updated": last_updated,
                    "hash": key,
                }
            )

        return results


class CorporationExporter(LedgerCSVExporter):
    """CSV Exporter for Corporation Ledger Data."""

    topic = "corporation-ledger"

    def __init__(
        self,
        entity_id: int,
        division_id: int = None,
        year: int = None,
        month: int = None,
    ):
        super().__init__()
        self.corporation_id = entity_id
        self.division_id = division_id
        self.year = year
        self.month = month

    @property
    def has_data(self) -> bool:
        """Check if there is data to export."""
        return CorporationWalletJournalEntry.objects.filter(
            division__corporation__corporation__corporation_id=self.corporation_id
        ).exists()

    def create_data_export(
        self,
        corporation_id: int = None,
        division_id: int = None,
        year: int = None,
        month: int = None,
    ) -> list[dict]:
        """Generate data export for corporation ledger."""
        report_corporation_id = (
            corporation_id if corporation_id is not None else self.corporation_id
        )
        report_division_id = (
            division_id if division_id is not None else self.division_id
        )
        report_year = year if year is not None else self.year
        report_month = month if month is not None else self.month

        try:
            corporation = CorporationAudit.objects.get(
                corporation__corporation_id=report_corporation_id
            )
            # Create CorporationData inside the task
            ledger_data = CorporationData(
                corporation=corporation,
                division_id=report_division_id,
                year=report_year,
                month=report_month,
            )
            return ledger_data.generate_data_export()
        except CorporationAudit.DoesNotExist as exc:
            raise ValueError("Corporation not found") from exc


class AllianceExporter(LedgerCSVExporter):
    """CSV Exporter for Alliance Ledger Data."""

    topic = "alliance-ledger"

    def __init__(self, entity_id: int, year: int = None, month: int = None):
        super().__init__()
        self.alliance_id = entity_id
        self.year = year
        self.month = month

    @property
    def has_data(self) -> bool:
        """Check if there is data to export."""
        corporations = CorporationAudit.objects.filter(
            corporation__alliance__alliance_id=self.alliance_id
        ).values_list("corporation__corporation_id", flat=True)

        if not corporations:
            return False

        return CorporationWalletJournalEntry.objects.filter(
            division__corporation__corporation__corporation_id__in=corporations
        ).exists()

    def create_data_export(
        self, alliance_id: int = None, year: int = None, month: int = None
    ) -> list[dict]:
        """Generate data export for alliance ledger."""
        report_alliance_id = (
            alliance_id if alliance_id is not None else self.alliance_id
        )
        report_year = year if year is not None else self.year
        report_month = month if month is not None else self.month

        try:
            alliance = EveAllianceInfo.objects.get(alliance_id=report_alliance_id)
            ledger_data = AllianceData(
                alliance=alliance,
                year=report_year,
                month=report_month,
                request=None,
            )
            return ledger_data.generate_data_export()
        except EveAllianceInfo.DoesNotExist as exc:
            raise ValueError("Alliance not found") from exc



# Django
from django.db import models
from django.db.models import Case, Count, Q, Value, When

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        """Returns a queryset of all corps visible to the user."""
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = []

            if user.has_perm("ledger.advanced_access"):
                # If the user has advanced access, return all corps from their characters
                corp_ids = char.character_ownership.user.character_ownerships.all().values_list(
                    "character__corporation_id", flat=True
                )
                queries.append(models.Q(corporation__corporation_id__in=corp_ids))

            logger.debug("%s queries for User %s.", len(queries), user)

            if len(queries) == 0:
                return self.none()

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def manage_to(self, user):
        """Return all corps that the user can manage."""
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = []

            if user.has_perm("ledger.corp_audit_manager"):
                # If the user has corp management access, return all corps from their characters
                corp_ids = char.character_ownership.user.character_ownerships.all().values_list(
                    "character__corporation_id", flat=True
                )
                queries.append(models.Q(corporation__corporation_id__in=corp_ids))

            logger.debug("%s queries for User %s.", len(queries), user)

            if len(queries) == 0:
                return self.none()

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def annotate_total_update_status_user(self, user):
        """Get the total update status for the given user."""
        # TODO: implement this method if needed
        return user

    def annotate_total_update_status(self):
        """Get the total update status."""
        # pylint: disable=import-outside-toplevel
        # AA Ledger
        from ledger.models.corporationaudit import CorporationAudit

        sections = CorporationAudit.UpdateSection.get_sections()
        num_sections_total = len(sections)
        qs = (
            self.annotate(
                num_sections_total=Count(
                    "ledger_corporation_update_status",
                    filter=Q(ledger_corporation_update_status__section__in=sections),
                )
            )
            .annotate(
                num_sections_ok=Count(
                    "ledger_corporation_update_status",
                    filter=Q(
                        ledger_corporation_update_status__section__in=sections,
                        ledger_corporation_update_status__is_success=True,
                    ),
                )
            )
            .annotate(
                num_sections_failed=Count(
                    "ledger_corporation_update_status",
                    filter=Q(
                        ledger_corporation_update_status__section__in=sections,
                        ledger_corporation_update_status__is_success=False,
                    ),
                )
            )
            .annotate(
                num_sections_token_error=Count(
                    "ledger_corporation_update_status",
                    filter=Q(
                        ledger_corporation_update_status__section__in=sections,
                        ledger_corporation_update_status__has_token_error=True,
                    ),
                )
            )
            # pylint: disable=no-member
            .annotate(
                total_update_status=Case(
                    When(
                        active=False,
                        then=Value(CorporationAudit.UpdateStatus.DISABLED),
                    ),
                    When(
                        num_sections_token_error=1,
                        then=Value(CorporationAudit.UpdateStatus.TOKEN_ERROR),
                    ),
                    When(
                        num_sections_failed__gt=0,
                        then=Value(CorporationAudit.UpdateStatus.ERROR),
                    ),
                    When(
                        num_sections_ok=num_sections_total,
                        then=Value(CorporationAudit.UpdateStatus.OK),
                    ),
                    When(
                        num_sections_total__lt=num_sections_total,
                        then=Value(CorporationAudit.UpdateStatus.INCOMPLETE),
                    ),
                    default=Value(CorporationAudit.UpdateStatus.IN_PROGRESS),
                )
            )
        )

        return qs


class CorporationAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)


CorporationAuditManager = CorporationAuditManagerBase.from_queryset(
    CorporationAuditQuerySet
)



"""
Corporation Audit
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.eveonline.providers import ObjectNotFound, provider
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.models.general import EveEntity

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@token_required(scopes=["publicData"])
@permission_required(["ledger.manage_access"])
def add_ally(request, token) -> HttpResponse:
    char = get_object_or_404(EveCharacter, character_id=token.character_id)
    try:
        ally = EveAllianceInfo.objects.get(alliance_id=char.alliance_id)

        msg = _("{alliance_name} is already in the Ledger System").format(
            alliance_name=ally.alliance_name,
        )
        messages.info(request, msg)
    except EveAllianceInfo.DoesNotExist:
        try:
            ally_data = provider.get_alliance(char.alliance_id)
            ally, __ = EveAllianceInfo.objects.get_or_create(
                alliance_id=ally_data.id,
                defaults={
                    "alliance_name": ally_data.name,
                    "alliance_ticker": ally_data.ticker,
                    "executor_corp_id": ally_data.executor_corp_id,
                },
            )
            # Add/Update All Corporations to eveuniverse model
            ally.populate_alliance()

            # Add the alliance to the EveEntity model
            EveEntity.objects.get_or_create(
                eve_id=ally.alliance_id,
                defaults={
                    "name": ally.alliance_name,
                    "category": "alliance",
                },
            )
            msg = _("{alliance_name} successfully added to Ledger").format(
                alliance_name=ally.alliance_name,
            )
            messages.success(request, msg)
        except ObjectNotFound:
            msg = _("Failed to fetch Alliance data for {alliance_name}").format(
                alliance_name=char.alliance_name,
            )
            messages.warning(request, msg)
            return redirect("ledger:alliance_ledger", alliance_id=char.alliance_id)
    return redirect("ledger:alliance_ledger", alliance_id=ally.alliance_id)



# Django
from django.utils import timezone

# AA Ledger
from ledger.models.events import Events


def create_event(char_ledger: bool, **kwargs) -> Events:
    """Create a Event for a Corporation"""
    params = {
        "char_ledger": char_ledger,
    }
    params.update(kwargs)
    events = Events(**params)
    events.save()
    return events


def create_event_1_day(
    char_ledger: bool, date_start: timezone.datetime, **kwargs
) -> Events:
    """Create a Event for a Corporation"""
    params = {
        "char_ledger": char_ledger,
        "date_start": date_start,
        "date_end": date_start + timezone.timedelta(days=1),
    }
    params.update(kwargs)
    events = Events(**params)
    events.save()
    return events



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import Mock, patch

# Django
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse

# AA Ledger
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.generate_planets import (
    _planetary_data,
    create_character_planet,
    create_character_planet_details,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views.character import planetary

PLANETARY_PATH = "ledger.views.character.planetary"


class TestViewSwitchAlarm(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.planet_params = {
            "upgrade_level": 5,
            "num_pins": 5,
        }

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.user_nopermission, cls.character_ownership_no_pm = (
            create_user_from_evecharacter_with_access(1002)
        )
        cls.audit = add_characteraudit_character_to_user(cls.user, 1001)
        cls.planetary = create_character_planet(cls.audit, 4001, **cls.planet_params)
        cls.planetarydetails = create_character_planet_details(
            cls.planetary, **_planetary_data
        )

    def test_switch_alarm(self):
        character_id = self.audit.eve_character.character_id
        form_data = {
            "character_id": character_id,
            "planet_id": 4001,
            "confirm": "yes",
        }

        request = self.factory.post(
            reverse("ledger:switch_alarm"),
            data=form_data,
        )
        request.user = self.user

        response = planetary.switch_alarm(request)

        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_switch_alarm_all(self):
        form_data = {
            "character_id": 0,
            "planet_id": 0,
            "confirm": "yes",
        }

        request = self.factory.post(
            reverse("ledger:switch_alarm"),
            data=form_data,
        )
        request.user = self.user

        response = planetary.switch_alarm(request)

        self.assertEqual(response.status_code, HTTPStatus.OK)

    @patch(PLANETARY_PATH + ".CharacterPlanetDetails.objects.filter")
    def test_switch_alarm_doesnotexist(self, mock_planetdetails_filter):
        form_data = {
            "character_id": 1001,
            "planet_id": 0,
            "confirm": "yes",
        }

        mock_queryset = Mock()
        mock_queryset.exists.return_value = False
        mock_planetdetails_filter.return_value = mock_queryset

        request = self.factory.post(
            reverse("ledger:switch_alarm"),
            data=form_data,
        )
        request.user = self.user

        response = planetary.switch_alarm(request)

        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)
        mock_planetdetails_filter.assert_called_once()

    def test_switch_alarm_no_permission(self):
        form_data = {
            "character_id": 1001,
            "planet_id": 4001,
            "confirm": "yes",
        }
        request = self.factory.post(
            reverse("ledger:switch_alarm"),
            data=form_data,
        )
        request.user = self.user_nopermission

        response = planetary.switch_alarm(request)

        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_switch_alarm_invalid(self):
        request = self.factory.post(
            reverse("ledger:switch_alarm"),
            data=None,
        )
        request.user = self.user

        response = planetary.switch_alarm(request)

        self.assertEqual(response.status_code, HTTPStatus.BAD_REQUEST)



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Ledger
from ledger import __version__


class LedgerConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    name = "ledger"
    label = "ledger"
    verbose_name = f"Ledger v{__version__}"

    def ready(self):
        """Ready"""
        # pylint: disable=import-outside-toplevel, unused-import
        from . import checks



"""Shared ESI client for Ledger."""

# Alliance Auth
from esi.openapi_clients import ESIClientProvider

# AA Ledger
from ledger import (
    __app_name_useragent__,
    __character_operations__,
    __corporation_operations__,
    __esi_compatibility_date__,
    __github_url__,
    __title__,
    __universe_operations__,
    __version__,
)

esi = ESIClientProvider(
    compatibility_date=__esi_compatibility_date__,
    ua_appname=__app_name_useragent__,
    ua_version=__version__,
    ua_url=__github_url__,
    operations=__corporation_operations__
    + __character_operations__
    + __universe_operations__,
)



"""PvE Views"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__

# Ledger
from ledger.helpers.core import add_info_to_context
from ledger.tasks import clear_all_etags, update_all_characters, update_all_corporations

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("ledger.basic_access")
def index(request):
    """Index View"""
    context = {
        "title": "Ledger",
    }
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:character_ledger",
        character_id=request.user.profile.main_character.character_id,
    )


@login_required
@permission_required("ledger.basic_access")
def admin(request):
    if not request.user.is_superuser:
        messages.error(request, _("You do not have permission to access this page."))
        return redirect("ledger:index")

    if request.method == "POST":
        force_refresh = False
        if request.POST.get("force_refresh", False):
            force_refresh = True
        if request.POST.get("run_clear_etag"):
            messages.info(request, _("Queued Clear All ETags"))
            clear_all_etags.apply_async(priority=1)
        if request.POST.get("run_char_updates"):
            messages.info(request, _("Queued Update All Characters"))
            update_all_characters.apply_async(
                kwargs={"force_refresh": force_refresh}, priority=7
            )
        if request.POST.get("run_corp_updates"):
            messages.info(request, _("Queued Update All Corporations"))
            update_all_corporations.apply_async(
                kwargs={"force_refresh": force_refresh}, priority=7
            )
    return render(request, "ledger/admin.html")



# Generated by Django 4.2.25 on 2025-10-17 05:40

# Django
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0016_characterminingledger_price_per_unit"),
    ]

    operations = [
        migrations.AlterField(
            model_name="eveentity",
            name="category",
            field=models.CharField(
                choices=[
                    ("alliance", "Alliance"),
                    ("corporation", "Corporation"),
                    ("character", "Character"),
                ],
                max_length=32,
                verbose_name="Category",
            ),
        ),
        migrations.AlterField(
            model_name="eveentity",
            name="eve_id",
            field=models.IntegerField(
                primary_key=True,
                serialize=False,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="ID",
            ),
        ),
        migrations.AlterField(
            model_name="eveentity",
            name="name",
            field=models.CharField(max_length=254, verbose_name="Name"),
        ),
    ]



"""
App Settings
"""

# Standard Library
import sys

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting

IS_TESTING = sys.argv[1:2] == ["test"]

# Set Test Mode True or False

# Set Naming on Auth Hook
LEDGER_APP_NAME = clean_setting("LEDGER_APP_NAME", "Ledger")

# zKillboard - https://zkillboard.com/
EVE_BASE_URL = "https://esi.evetech.net/"
EVE_API_URL = "https://esi.evetech.net/latest/"
EVE_BASE_URL_REGEX = r"^http[s]?:\/\/esi.evetech\.net\/"

# fuzzwork
FUZZ_BASE_URL = "https://www.fuzzwork.co.uk/"
FUZZ_API_URL = "https://www.fuzzwork.co.uk/api/"
FUZZ_BASE_URL_REGEX = r"^http[s]?:\/\/(www\.)?fuzzwork\.co\.uk\/"

# Global timeout for tasks in seconds to reduce task accumulation during outages.
LEDGER_TASKS_TIME_LIMIT = clean_setting("LEDGER_TASKS_TIME_LIMIT", 600)

LEDGER_STALE_TYPES = clean_setting(
    "LEDGER_STALE_TYPES",
    {
        "wallet_journal": 30,
        "wallet_division_names": 30,
        "wallet_division": 30,
        "mining_ledger": 30,
        "planets": 30,
        "planets_details": 30,
    },
)

# Mining Price Calculation
LEDGER_USE_COMPRESSED = clean_setting("LEDGER_USE_COMPRESSED", True)
LEDGER_PRICE_PERCENTAGE = clean_setting("LEDGER_PRICE_PERCENTAGE", 0.9)

# Ledger Cache System
LEDGER_CACHE_STALE = 60 * 60 * 168  # 168 hours
LEDGER_CACHE_KEY = "LEDGER"
LEDGER_CACHE_ENABLED = True



"""
Corporation Audit Model
"""

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.managers.corporation_audit_manager import CorporationAuditManager
from ledger.managers.corporation_journal_manager import (
    CorporationDivisionManager,
    CorporationWalletManager,
)
from ledger.models.characteraudit import WalletJournalEntry
from ledger.models.general import (
    AuditBase,
    UpdateStatus,
)
from ledger.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationAudit(AuditBase):
    """A model to store corporation information."""

    class UpdateSection(models.TextChoices):
        WALLET_DIVISION_NAMES = "wallet_division_names", _("Divisions Names")
        WALLET_DIVISION = "wallet_division", _("Divisions")
        WALLET_JOURNAL = "wallet_journal", _("Wallet Journal")

        @classmethod
        def get_sections(cls) -> list[str]:
            """Return list of section values."""
            return [choice.value for choice in cls]

        @property
        def method_name(self) -> str:
            """Return method name for this section."""
            return f"update_{self.value}"

    corporation_name = models.CharField(max_length=100, null=True, default=None)

    active = models.BooleanField(default=True)

    corporation = models.OneToOneField(
        EveCorporationInfo,
        on_delete=models.CASCADE,
        related_name="ledger_corporationaudit",
    )

    objects = CorporationAuditManager()

    def __str__(self) -> str:
        try:
            return f"{self.corporation.corporation_name} ({self.id})"
        except AttributeError:
            return f"{self.corporation} ({self.id})"

    class Meta:
        default_permissions = ()
        permissions = (
            ("corp_audit_manager", "Has Access to own Corporations."),
            ("corp_audit_admin_manager", "Has access to all Corporations."),
        )

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-search.search_structures.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_corporation_roles.v1",
            # Mining
            "esi-industry.read_corporation_mining.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-markets.read_corporation_orders.v1",
            "esi-industry.read_corporation_jobs.v1",
            "esi-corporations.read_divisions.v1",
        ]

    @property
    def get_status(self) -> UpdateStatus:
        """Get the status of this corporation."""
        if self.active is False:
            return self.UpdateStatus.DISABLED

        qs = CorporationAudit.objects.filter(pk=self.pk).annotate_total_update_status()
        total_update_status = list(qs.values_list("total_update_status", flat=True))[0]
        return self.UpdateStatus(total_update_status)

    @property
    def update_status(self):
        return self.ledger_corporation_update_status

    def get_token(self, scopes, req_roles) -> Token:
        """Get the token for this corporation."""
        if "esi-characters.read_corporation_roles.v1" not in scopes:
            scopes.append("esi-characters.read_corporation_roles.v1")

        char_ids = EveCharacter.objects.filter(
            corporation_id=self.corporation.corporation_id
        ).values("character_id")

        tokens = Token.objects.filter(character_id__in=char_ids).require_scopes(scopes)

        for token in tokens:
            try:
                roles = esi.client.Character.GetCharactersCharacterIdRoles(
                    character_id=token.character_id, token=token
                ).result(force_refresh=True)

                has_roles = False
                for role in roles.roles:
                    if role in req_roles:
                        has_roles = True

                if has_roles:
                    return token
            except TokenError as e:
                logger.error(
                    "Token ID: %s (%s)",
                    token.pk,
                    e,
                )
        return False

    def update_wallet_division_names(self, force_refresh: bool) -> None:
        return self.ledger_corporation_division.update_or_create_esi_names(
            self, force_refresh=force_refresh
        )

    def update_wallet_division(self, force_refresh: bool) -> None:
        return self.ledger_corporation_division.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def update_wallet_journal(self, force_refresh: bool) -> None:
        return CorporationWalletJournalEntry.objects.update_or_create_esi(
            self, force_refresh=force_refresh
        )


class CorporationWalletDivision(models.Model):
    objects = CorporationDivisionManager()

    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_division",
    )
    name = models.CharField(max_length=100, null=True, default=None)
    balance = models.DecimalField(max_digits=20, decimal_places=2)
    division_id = models.IntegerField()

    class Meta:
        default_permissions = ()


class CorporationWalletJournalEntry(WalletJournalEntry):
    division = models.ForeignKey(
        CorporationWalletDivision,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_journal",
    )

    objects = CorporationWalletManager()

    def __str__(self):
        return f"Corporation Wallet Journal: RefType: {self.ref_type} - {self.first_party.name} -> {self.second_party.name}: {self.amount} ISK"

    @classmethod
    def get_visible(cls, user):
        corps_vis = CorporationAudit.objects.visible_to(user)
        return cls.objects.filter(division__corporation__in=corps_vis)


class CorporationUpdateStatus(UpdateStatus):
    """A Model to track the status of the last update."""

    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_update_status",
    )
    section = models.CharField(
        max_length=32, choices=CorporationAudit.UpdateSection.choices, db_index=True
    )

    def __str__(self) -> str:
        return f"{self.corporation} - {self.section} - {self.is_success}"



# Third Party
from ninja import NinjaAPI

# Django
from django.db.models import Q
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.api import schema
from ledger.api.api_helper.planetary_helper import (
    generate_progressbar,
    get_facilities_info,
)
from ledger.api.helpers import get_alts_queryset, get_character_or_none
from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class LedgerPlanetaryApiEndpoints:
    tags = ["CharacterPlanet"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "character/{character_id}/planetary/{planet_id}/",
            response={200: list[schema.CharacterPlanet], 403: str},
            tags=self.tags,
        )
        def get_planetary(request, character_id: int, planet_id: int):
            is_all = False
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
                is_all = True

            perm, character = get_character_or_none(request, character_id)

            if not perm:
                return 403, str(_("Permission Denied"))

            if is_all:
                characters = get_alts_queryset(character)
            else:
                characters = [character]

            filters = Q(character__eve_character__in=characters)
            if not planet_id == 0:
                filters &= Q(planet__id=planet_id)

            planets = CharacterPlanet.objects.filter(filters).select_related(
                "planet", "character"
            )

            output = []

            for p in planets:
                output.append(
                    {
                        "character_id": p.character.eve_character.character_id,
                        "character_name": p.character.eve_character.character_name,
                        "planet": p.planet.name,
                        "planet_id": p.planet.id,
                        "upgrade_level": p.upgrade_level,
                        "num_pins": p.num_pins,
                        "last_update": p.last_update,
                    }
                )
            return output

        @api.get(
            "character/{character_id}/planetary/{planet_id}/details/",
            response={200: list, 403: str},
            tags=self.tags,
        )
        # pylint: disable=too-many-locals
        def get_planetarydetails(request, character_id: int, planet_id: int):
            singleview = request.GET.get("single", False)
            perm, character = get_character_or_none(request, character_id)

            if not perm:
                return 403, str(_("Permission Denied"))

            if not singleview:
                characters = get_alts_queryset(character)
            else:
                characters = [character]

            filters = Q(planet__character__in=characters)
            if not planet_id == 0:
                filters &= Q(planet__planet__id=planet_id)

            planets = (
                CharacterPlanetDetails.objects.filter(filters)
                .prefetch_related(
                    "planet",
                    "planet__planet",
                    "planet__character",
                )
                .select_related("planet__planet")
            )

            output = []

            for p in planets:
                products_types = p.allocate_products()
                extracts = p.allocate_extracts()

                products = {
                    "raw": extracts,
                    "processed": products_types,
                }

                output.append(
                    {
                        "character_id": p.planet.character.eve_character.character_id,
                        "character_name": p.planet.character.eve_character.character_name,
                        "planet": p.planet.planet.name,
                        "planet_id": p.planet.planet.id,
                        "planet_type_id": p.planet.planet.eve_type.id,
                        "upgrade_level": p.planet.upgrade_level,
                        "expiry_date": p.get_planet_expiry_date(),
                        "expired": p.is_expired,
                        "alarm": p.notification,
                        "percentage": generate_progressbar(
                            p.allocate_overall_progress()
                        ),
                        "products": products,
                        "storage": p.get_storage_info(),
                        "facility": p.facilitys,
                        "last_update": p.planet.last_update,
                    }
                )
            return output

        @api.get(
            "character/{character_id}/planetary/{planet_id}/factory/",
            response={200: dict, 403: str},
            tags=self.tags,
        )
        def get_factory_info(request, character_id: int, planet_id: int):
            # Get the character
            perm, character = get_character_or_none(request, character_id)

            if not perm:
                return 403, str(_("Permission Denied"))

            filters = Q(planet__character=character)
            if planet_id != 0:
                filters &= Q(planet__planet__id=planet_id)

            try:
                planet = CharacterPlanetDetails.objects.get(filters)
                storage = planet.get_storage_info()
                facilities = get_facilities_info(planet)

                output = {
                    "title": _("Factory Information"),
                    "character_id": planet.planet.character.eve_character.character_id,
                    "character_name": planet.planet.character.eve_character.character_name,
                    "planet_id": planet.planet.planet.id,
                    "planet_name": planet.planet.planet.name,
                    "facilities": facilities,
                    "storage": storage,
                    "last_update": planet.planet.last_update,
                }
            except CharacterPlanetDetails.DoesNotExist:
                output = {
                    "title": _("Planet not Found"),
                }

            context = {
                "character": output,
                "mode": "factory",
            }

            return render(
                request,
                "ledger/partials/modal/view_factory.html",
                context,
            )

        @api.get(
            "character/{character_id}/planetary/{planet_id}/extractor/",
            response={200: dict, 403: str},
            tags=self.tags,
        )
        def get_extractor_info(request, character_id: int, planet_id: int):
            perm, character = get_character_or_none(request, character_id)

            if not perm:
                return 403, str(_("Permission Denied"))

            filters = Q(planet__character=character)
            if not planet_id == 0:
                filters &= Q(planet__planet__id=planet_id)

            try:
                planet = CharacterPlanetDetails.objects.get(filters)
                output = {
                    "title": _("Extractor Information"),
                    "character_id": planet.planet.character.eve_character.character_id,
                    "character_name": planet.planet.character.eve_character.character_name,
                    "planet_id": planet.planet.planet.id,
                    "planet_name": planet.planet.planet.name,
                    "extractors": planet.get_extractors_info(),
                    "last_update": planet.planet.last_update,
                }
            except CharacterPlanetDetails.DoesNotExist:
                output = {
                    "title": _("Planet not Found"),
                }
                return output

            context = {
                "character": output,
                "mode": "extractor",
            }

            return render(
                request,
                "ledger/partials/modal/view_extractor.html",
                context,
            )



"""PvE Views"""

# Standard Library
from decimal import Decimal
from typing import Any

# Django
from django.db.models import DecimalField, Q, QuerySet, Sum
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__
from ledger.constants import NPC_ENTITIES
from ledger.helpers.billboard import BillboardSystem
from ledger.helpers.core import LedgerCore, LedgerEntity
from ledger.helpers.ref_type import RefTypeManager
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CorporationData(LedgerCore):
    """Class to hold character data for the ledger."""

    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        corporation: CorporationAudit,
        division_id: int = None,
        year: int = None,
        month: int = None,
        day: int = None,
        section: str = None,
    ):
        super().__init__(year, month, day)
        self.corporation = corporation
        self.entity_id = corporation.corporation.corporation_id
        self.division_id = division_id
        self.section = section
        self.auth_char_ids = self.auth_character_ids
        self.billboard = BillboardSystem()
        self.queryset = (
            self._get_journal_queryset()
        )  # Prepare Queryset and auxiliary data

    @property
    def has_data(self) -> bool:
        """Return True if there is data for the corporation in the selected date range."""
        return self.queryset.exists()

    @property
    def activity_years(self) -> list[int]:
        """Return a list of years with data for the corporation."""
        return (
            CorporationWalletJournalEntry.objects.filter(
                division__corporation=self.corporation
            )
            .values_list("date__year", flat=True)
            .distinct()
            .order_by("-date__year")
        )

    @property
    def divisions(self) -> QuerySet:
        """Return a list of divisions for the corporation."""
        return CorporationWalletDivision.objects.filter(
            corporation=self.corporation
        ).order_by("division_id")

    def _get_journal_queryset(self) -> QuerySet[CorporationWalletJournalEntry]:
        """Return the base queryset filtered by the current date range and corporation division."""
        if self.division_id is not None:
            return CorporationWalletJournalEntry.objects.filter(
                self.filter_date,
                division__corporation=self.corporation,
                division=self.division_id,
            ).exclude(
                first_party_id=self.corporation.corporation.corporation_id,
                second_party_id=self.corporation.corporation.corporation_id,
            )
        return CorporationWalletJournalEntry.objects.filter(
            self.filter_date,
            division__corporation=self.corporation,
        ).exclude(
            first_party_id=self.corporation.corporation.corporation_id,
            second_party_id=self.corporation.corporation.corporation_id,
        )

    def _compute_entities(
        self, journal: QuerySet[CorporationWalletJournalEntry]
    ) -> set:
        """Return a set of all entity IDs (first and second parties) present in the current journal."""
        return set(journal.values_list("second_party_id", flat=True)) | set(
            journal.values_list("first_party_id", flat=True)
        )

    # pylint: disable=duplicate-code
    def _compute_journal_values(
        self, journal: QuerySet[CorporationWalletJournalEntry]
    ) -> QuerySet[dict[str, Any]]:
        """Return the journal values for the current journal."""
        return journal.values(
            "first_party_id", "second_party_id", "pk", "ref_type"
        ).annotate(
            bounty=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.BOUNTY_PRIZES),
                output_field=DecimalField(),
            ),
            ess=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.ESS_TRANSFER),
                output_field=DecimalField(),
            ),
            costs=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__lt=0),
                output_field=DecimalField(),
            ),
            miscellaneous=Sum(
                "amount",
                filter=Q(ref_type__in=RefTypeManager.all_ref_types(), amount__gt=0),
                output_field=DecimalField(),
            ),
        )

    def create_entity_data(
        self,
        entity: LedgerEntity,
        alts: EveCharacter = None,
    ) -> dict:
        """Create the URL for entity details based on the view type."""
        ids = (
            list(alts.values_list("character__character_id", flat=True))
            if alts is not None
            else [entity.entity_id]
        )

        # Create Alts Dictionary
        alts_dict = {}
        if alts is not None:
            for alt in alts:
                alts_dict[alt.character.character_id] = alt.character.character_name

            # Remove the main character from the alts dictionary only one entry
            if len(alts_dict) == 1:
                alts_dict.pop(entity.entity_id, None)

        used_pks = set()
        bounty = Decimal(0)
        ess = Decimal(0)
        miscellaneous = Decimal(0)
        costs = Decimal(0)

        for pk, rows in list(self.entries.items()):
            for row in rows:
                if row["first_party_id"] in ids or row["second_party_id"] in ids:
                    if RefTypeManager.special_cases(
                        row, ids=ids, account_char_ids=self.auth_char_ids
                    ):
                        continue
                    bounty += row.get("bounty") or Decimal(0)
                    ess += row.get("ess") or Decimal(0)
                    miscellaneous += row.get("miscellaneous") or Decimal(0)
                    costs += row.get("costs") or Decimal(0)
                    used_pks.add(pk)

        # Remove Used Pks from Entries
        # This is to prevent the entries from being used in the future
        for pk in used_pks:
            self.entries.pop(pk, None)

        total = sum([bounty, ess, miscellaneous, costs])

        if total == 0:
            return None

        char_data = {
            "entity": entity,
            "alts": alts_dict,
            "ledger": {
                "bounty": bounty,
                "ess": ess,
                "miscellaneous": miscellaneous,
                "costs": costs,
                "total": total,
            },
            "type": entity.type,
        }

        return char_data

    def generate_ledger_data(self) -> dict:
        """
        Generate the ledger data for the corporation.

        This method processes the journal entries, builds the ledger data,
        and prepares the context for rendering the corporation ledger view.
        """
        # Compute all entities in the journal
        self.entities = self._compute_entities(self.queryset)
        # Compute journal values
        journal = self._compute_journal_values(self.queryset)

        # Caching
        ledger_hash = self.get_ledger_journal_hash(journal.values_list("pk"))
        cache_key = f"{self.corporation.corporation.corporation_id}_{self.division_id}"

        # Get Cached Data if available
        ledger, finished_entities = self.get_cache_ledger(
            ledger_hash=ledger_hash, cache_key=cache_key
        )

        if finished_entities is False or ledger is False:
            # Build the entries from the journal
            self.entries = {}
            for row in journal:
                self.entries.setdefault(row["pk"], []).append(row)

            # Process Auth Accounts first
            ledger, finished_entities = self._process_auth_accounts()
            # Process remaining entities
            ledger, finished_entities = self._process_remaining_entities(
                ledger, finished_entities
            )
            # Process corporation entity last to ensure it's always included
            self._handle_entity(
                ledger=ledger,
                entity_id=self.corporation.corporation.corporation_id,
                corporation_obj=self.corporation.corporation,
            )

            # Create Cache
            self.set_cache_ledger(
                ledger_hash=ledger_hash,
                cache_key=cache_key,
                ledger=ledger,
                finished_entities=finished_entities,
            )

        # Finalize the billboard for the ledger.
        self.billboard.change_view(self.get_view_mode())
        self.create_rattingbar(journal=journal, entities_ids=list(finished_entities))
        self.create_chord(ledger)
        return ledger

    def generate_data_export(self) -> dict:
        """Generate the data export for the corporation."""
        # Compute all entities in the journal
        self.entities = self._compute_entities(self.queryset)
        # Compute journal values
        journal = self._compute_journal_values(self.queryset)

        # Build the entries from the journal
        self.entries = {}
        for row in journal:
            self.entries.setdefault(row["pk"], []).append(row)

        # Process Auth Accounts first
        ledger, finished_entities = self._process_auth_accounts()
        # Process remaining entities
        ledger, finished_entities = self._process_remaining_entities(
            ledger, finished_entities
        )
        # Process corporation entity last to ensure it's always included
        self._handle_entity(
            ledger=ledger,
            entity_id=self.corporation.corporation.corporation_id,
            corporation_obj=self.corporation.corporation,
        )
        return ledger

    # pylint: disable=too-many-arguments
    def _handle_entity(
        self,
        ledger: list,
        entity_id: int,
        character_obj=None,
        corporation_obj=None,
        alts=None,
        add_finished: bool = True,
        finished_ids=None,
    ) -> set:
        """Create entity object, add to ledger if it has data and return IDs to mark finished.

        - ledger: list to append to
        - entity_id: numeric id
        - character_obj / corporation_obj: optional objects to attach to LedgerEntity
        - alts: optional alts queryset passed to create_entity_data
        - add_finished: whether to return ids that should be added to finished_entities
        - finished_ids: explicit IDs (set or iterable) to mark finished (used for accounts)
        """
        details_url = self.create_url(
            viewname="corporation_details",
            corporation_id=self.corporation.corporation.corporation_id,
            entity_id=entity_id,
            section="single",
        )
        entity_obj = LedgerEntity(
            entity_id,
            character_obj=character_obj,
            corporation_obj=corporation_obj,
            details_url=details_url,
        )

        # Create the character data for the ledger
        char_data = self.create_entity_data(entity=entity_obj, alts=alts)

        # Only add to ledger if there is data
        if char_data is None:
            return set()

        ledger.append(char_data)

        if not add_finished:
            return set()

        if finished_ids is not None:
            return set(finished_ids)

        return {entity_id}

    def _process_auth_accounts(self):
        """Process Auth Account information for the ledger."""
        ledger = []
        finished_entities = set()
        for account in self.auth_accounts:
            alts = account.user.character_ownerships.all()
            existing_alts = set(
                alts.values_list("character__character_id", flat=True)
            ).intersection(self.entities)
            alts = alts.filter(character__character_id__in=existing_alts)
            if not existing_alts:
                continue
            finished_entities.update(
                self._handle_entity(
                    ledger,
                    account.main_character.character_id,
                    character_obj=account.main_character,
                    alts=alts,
                    finished_ids=existing_alts,
                )
            )
        return ledger, finished_entities

    def _process_remaining_entities(self, ledger, finished_entities: set):
        """Process remaining entities for the ledger."""
        remaining_entities = self.entities - finished_entities

        # If no remaining entities, return finished_entities
        if not remaining_entities:
            return ledger, finished_entities

        for entity_id in remaining_entities:
            if entity_id in NPC_ENTITIES:
                continue
            if entity_id == self.corporation.corporation.corporation_id:
                continue
            finished_entities.update(self._handle_entity(ledger, entity_id))
        return ledger, finished_entities

    def create_rattingbar(
        self,
        journal: QuerySet[CorporationWalletJournalEntry],
        entities_ids: list = None,
    ):
        """Create the ratting bar for the view."""
        if not entities_ids:
            return

        # Create the timeline for the ratting bar
        rattingbar_timeline = self.billboard.create_timeline(journal)
        # Annotate the timeline with the relevant data
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        # Generate the XY series for the ratting bar
        self.billboard.create_or_update_results(rattingbar)
        series, categories = self.billboard.generate_xy_series()
        if series and categories:
            # Create the ratting bar chart
            self.billboard.create_xy_chart(
                title=_("Ratting Bar"), categories=categories, series=series
            )

    def create_chord(self, ledger_data: list[dict]):
        """Create the chord chart for the view."""
        if not ledger_data:
            return

        for entry in ledger_data:
            entity_name = entry["entity"].entity_name
            ledger = entry["ledger"]
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Bounty (Wallet)"),
                value=ledger.get("bounty", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("ESS (Wallet)"),
                value=ledger.get("ess", 0),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Costs (Wallet)"),
                value=abs(ledger.get("costs", 0)),
            )
            self.billboard.chord_add_data(
                chord_from=entity_name,
                chord_to=_("Miscellaneous (Wallet)"),
                value=abs(ledger.get("miscellaneous", 0)),
            )
        self.billboard.chord_handle_overflow()



# Standard Library
from unittest.mock import patch

# Django
from django.test import TestCase

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA Ledger
from ledger.models.general import EveEntity
from ledger.tests.testdata.esi_stub import esi_client_stub_openapi
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.managers.general_manager"


class TestGeneralManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.manager = EveEntity.objects

    @patch("ledger.managers.general_manager.esi")
    def test_get_or_create_esi_existing(self, mock_esi):
        entity = EveEntity.objects.get(eve_id=1001)
        result, created = self.manager.get_or_create_esi(eve_id=1001)
        self.assertEqual(result, entity)
        self.assertFalse(created)

    @patch("ledger.managers.general_manager.esi")
    def test_get_or_create_esi_new(self, mock_esi):
        mock_esi.client = esi_client_stub_openapi
        result, created = self.manager.get_or_create_esi(eve_id=9996)
        self.assertEqual(result.eve_id, 9996)
        self.assertEqual(result.name, "Create Character")
        self.assertTrue(created)

    @patch("ledger.managers.general_manager.esi")
    def test_create_bulk_from_esi(self, mock_esi):
        mock_esi.client = esi_client_stub_openapi
        result = self.manager.create_bulk_from_esi([9997, 9998])
        self.assertTrue(result)
        self.assertTrue(EveEntity.objects.filter(eve_id=9997).exists())
        self.assertTrue(EveEntity.objects.filter(eve_id=9998).exists())

    @patch("ledger.managers.general_manager.esi")
    def test_update_or_create_esi(self, mock_esi):
        mock_esi.client = esi_client_stub_openapi
        result, created = self.manager.update_or_create_esi(eve_id=9999)
        self.assertEqual(result.eve_id, 9999)
        self.assertEqual(result.name, "New Test Character")
        self.assertTrue(created)



"""
Corporation Audit
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as trans

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Ledger
from ledger import __title__, tasks
from ledger.models.corporationaudit import CorporationAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@token_required(scopes=CorporationAudit.get_esi_scopes())
@permission_required(["ledger.manage_access"])
def add_corp(request, token) -> HttpResponse:
    char = get_object_or_404(EveCharacter, character_id=token.character_id)
    eve_corp, _ = EveCorporationInfo.objects.get_or_create(
        corporation_id=char.corporation_id,
        defaults={
            "member_count": 0,
            "corporation_ticker": char.corporation_ticker,
            "corporation_name": char.corporation_name,
        },
    )

    corp = CorporationAudit.objects.update_or_create(
        corporation=eve_corp,
        defaults={
            "corporation_name": eve_corp.corporation_name,
        },
    )[0]

    tasks.update_corporation.apply_async(
        args=[corp.pk], kwargs={"force_refresh": True}, priority=6
    )
    msg = trans("{corporation_name} successfully added/updated to Ledger").format(
        corporation_name=corp.corporation_name,
    )
    messages.info(request, msg)
    return redirect("ledger:corporation_ledger", corporation_id=eve_corp.corporation_id)



"""Initialize the app"""

__version__ = "1.0.3"
__title__ = "Ledger"

__package_name__ = "aa-ledger"
__app_name__ = "ledger"
__esi_compatibility_date__ = "2025-11-06"
__app_name_useragent__ = "AA-Ledger"

__github_url__ = f"https://github.com/Geuthur/{__package_name__}"

__character_operations__ = [
    "GetCharactersCharacterIdWalletJournal",
    "GetCharactersCharacterIdRoles",
    "GetCharactersCharacterIdPlanets",
    "GetCharactersCharacterIdPlanetsPlanetId",
    "GetCharactersCharacterIdMining",
]

__corporation_operations__ = [
    "GetCorporationsCorporationIdWallets",
    "GetCorporationsCorporationIdWalletsDivisionJournal",
    "GetCorporationsCorporationIdDivisions",
]

__universe_operations__ = ["PostUniverseNames"]



# Standard Library
from unittest.mock import PropertyMock, patch

# Django
from django.test import RequestFactory, TestCase
from django.utils import timezone

# AA Ledger
from ledger.models.characteraudit import CharacterAudit, CharacterUpdateStatus
from ledger.models.general import _NeedsUpdate
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_update_status,
    create_user_from_evecharacter_with_access,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "ledger.models.characteraudit"


class TestCharacterWalletJournalModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001,
        )
        cls.audit = add_characteraudit_character_to_user(
            cls.user, cls.character_ownership.character.character_id
        )
        sections = CharacterAudit.UpdateSection.get_sections()
        for section in sections:
            create_update_status(
                cls.audit,
                section=section,
                is_success=True,
                error_message="",
                has_token_error=False,
                last_run_at=timezone.now(),
                last_run_finished_at=timezone.now(),
                last_update_at=timezone.now(),
                last_update_finished_at=timezone.now(),
            )
        cls.update_status_wallet = CharacterUpdateStatus.objects.get(
            character=cls.audit,
            section=CharacterAudit.UpdateSection.WALLET_JOURNAL,
        )
        cls.update_status_mining_ledger = CharacterUpdateStatus.objects.get(
            character=cls.audit,
            section=CharacterAudit.UpdateSection.MINING_LEDGER,
        )

    def test_str(self):
        excepted_str = CharacterAudit.objects.get(id=self.audit.id)
        self.assertEqual(self.audit, excepted_str)

    def test_get_esi_scopes(self):
        self.assertEqual(
            self.audit.get_esi_scopes(),
            [
                # Mining Ledger
                "esi-industry.read_character_mining.v1",
                # Wallet / Market /  Contracts
                "esi-wallet.read_character_wallet.v1",
                "esi-contracts.read_character_contracts.v1",
                # Planetary Interaction
                "esi-planets.manage_planets.v1",
            ],
        )

    def test_get_status_states(self):
        update_status = self.update_status_wallet
        audit = self.audit

        self.assertEqual(
            self.audit.get_status,
            CharacterAudit.UpdateStatus.OK,
        )

        audit.active = False
        audit.save()

        self.assertEqual(
            self.audit.get_status,
            CharacterAudit.UpdateStatus.DISABLED,
        )

        audit.active = True
        audit.save()
        update_status.is_success = False
        update_status.has_token_error = True
        update_status.save()

        self.assertEqual(
            self.audit.get_status,
            CharacterAudit.UpdateStatus.TOKEN_ERROR,
        )

        audit.active = True
        audit.save()
        update_status.is_success = False
        update_status.has_token_error = False
        update_status.save()

        self.assertEqual(
            self.audit.get_status,
            CharacterAudit.UpdateStatus.ERROR,
        )

        update_status.delete()

        self.assertEqual(
            self.audit.get_status,
            CharacterAudit.UpdateStatus.INCOMPLETE,
        )

    def test_reset_has_token_error(self):
        audit = self.audit

        self.assertEqual(
            audit.reset_has_token_error(),
            False,
        )

        self.update_status_mining_ledger.has_token_error = True
        self.update_status_mining_ledger.save()

        self.assertEqual(
            audit.reset_has_token_error(),
            True,
        )

    def test_reset_update_status(self):
        # given
        audit = self.audit
        # when
        update_status = audit.reset_update_status("mining_ledger")
        # then
        self.assertEqual(
            update_status,
            CharacterUpdateStatus.objects.get(
                character=audit,
                section="mining_ledger",
            ),
        )



# Standard Library
from unittest.mock import MagicMock, Mock, patch

# Django
from django.core.cache import cache
from django.test import TestCase

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

MODULE_PATH = "ledger.helpers.etag"



"""TestView class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# AA Ledger
from ledger.tests.testdata.generate_characteraudit import (
    add_characteraudit_character_to_user,
    create_user_from_evecharacter_with_access,
)

# AA Skillfarm
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.views.character.character_ledger import character_delete

MODULE_PATH = "ledger.views.character.character_ledger"


class TestDeleteCharacterView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter_with_access(
            1001
        )
        cls.audit = add_characteraudit_character_to_user(cls.user, 1001)
        cls.no_audit_user, cls.character_ownership = (
            create_user_from_evecharacter_with_access(1002)
        )

    def test_delete_character(self):
        character_id = self.audit.eve_character.character_id

        request = self.factory.post(reverse("ledger:delete_char", args=[character_id]))
        request.user = self.user

        response = character_delete(request, character_id=character_id)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTrue(response_data["success"])
        self.assertEqual(response_data["message"], "Gneuten successfully deleted")

    def test_delete_character_no_permission(self):
        request = self.factory.post(reverse("ledger:delete_char", args=[1001]))
        request.user = self.no_audit_user

        response = character_delete(request, character_id=1001)

        response_data = json.loads(response.content)

        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Permission Denied")



"""Generate AllianceAuth test objects from allianceauth.json."""

# Standard Library
import json
from pathlib import Path

# Alliance Auth (External Libs)
from eveuniverse.models import EvePlanet

# AA Ledger
from ledger.models.characteraudit import CharacterAudit
from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails


def _load_planetary_data():
    with open(Path(__file__).parent / "planetary.json", encoding="utf-8") as fp:
        return json.load(fp)


_planetary_data = _load_planetary_data()


def create_character_planet(
    characteraudit: CharacterAudit, planet_id: int, **kwargs
) -> CharacterPlanet:
    """Create a CharacterPlanet from CharacterAudit and planet_id."""
    params = {
        "character": characteraudit,
        "planet": EvePlanet.objects.get(id=planet_id),
    }
    params.update(kwargs)
    planet = CharacterPlanet(**params)
    planet.save()
    return planet


def create_character_planet_details(
    characterplanet: CharacterPlanet, **kwargs
) -> CharacterPlanetDetails:
    """Create a CharacterPlanetDetails from CharacterPlanet."""
    params = {
        "character": characterplanet.character,
        "planet": characterplanet,
    }
    params.update(kwargs)
    planetdetails = CharacterPlanetDetails(**params)
    planetdetails.save()
    return planetdetails
